// WCS.java

package astroj;

import java.awt.*;
import java.util.*;

import ij.*;

/**
 * A simple WCS object for the world coordinate systems used in most astronomical FITS files.
 * Uses the FitsJ class methods.
 *
 * The processing follows the general guidelines in
 *	Greisen, E.W. & Calabretta M.R. 2002, Astron. & Astrophys. Vol. 395, 1061-1075
 *	Calabretta M.R. & Greisen, E.W. 2002, Astron. & Astrophys. Vol. 395, 1077-1122
 *
 * Modifications by K. Collins 7/2010
 * Fixed three issues related to proper RA and DEC calculations for the TAN projection
 * - The "RA---TAN-SIP" extension caused problems with the "-TAN" substring matching
 *        changed: int i = CTYPE[0].lastIndexOf("-");
 *                 projection = CTYPE[0].substring(i+1);
 *             to: projection = CTYPE[0].substring(5, 8);
 *
 * - Reversed the indices when loading the PC and CD Matrix coefficients
 *        changed: CD[k-1][i-1]  and PC[k-1][i-1]
 *             to: CD[i-1][k-1]  and PC[i-1][k-1]
 *
 * - Reversed the order of the arguments in one statement when calculating the TAN projection
 *        changed: s[0] = DEGRAD*Math.atan2(-x[1],x[0]);
 *             to: s[0] = DEGRAD*Math.atan2(x[0],-x[1]);
 *
 * Added code to support the SIP non-linear distortion corrections following
 *  SHUPE, D.L. ET AL. 2008
 *
 */
public class WCS
	{
	int NAXIS=0;
	int WCSAXES=0;
	int[] NAXES;
	double[] CDELT;
	double[] CRPIX;
	double[] CRVAL;
	double[][] PC;
	double[][] CD;
    double[][] A;
    double[][] B;
	String[] CTYPE;
	String[] CUNIT;
	double LONPOLE;
    int A_ORDER;
    int B_ORDER;

	public String coordsys;
	String projection;
	boolean useCD=false;
    boolean useSIPA=false;
    boolean useSIPB=false;
	boolean hasRADEC=false;
    boolean typeContainsSIP = false;

	boolean enoughInfo=false;
	public String logInfo = new String("");

	public static String PREFS_LONGITUDES = new String("wcs.preflongs");
	public static String PREFS_LATITUDES = new String("wcs.preflats");
	public static String PREFS_CRPIX1 = new String("wcs.crpix1");
	public static String PREFS_CRPIX2 = new String("wcs.crpix2");
	public static String PREFS_CD1_1 = new String("wcs.cd11");
	public static String PREFS_CD1_2 = new String("wcs.cd12");
	public static String PREFS_CD2_1 = new String("wcs.cd21");
	public static String PREFS_CD2_2 = new String("wcs.cd22");
	public static String PREFS_NPIX1 = new String("wcs.npix1");
	public static String PREFS_NPIX2 = new String("wcs.npix2");

	public WCS (int naxis)
		{
		initialize(naxis);
		}

	public WCS (ImagePlus img)
		{
		NAXIS=0;
		String[] hdr = FitsJ.getHeader (img);
		if (hdr == null) return;
		process(img.getShortTitle(),img.getWidth(),img.getHeight(),img.getStackSize(),hdr);
		}

	public WCS (String[] hdr)
		{
		NAXIS=0;
		if (hdr == null) return;
		process(null,-1,-1,-1, hdr);
		}

	protected void process (String title, int nx, int ny, int nz, String[] hdr)
		{
		// FITS HEADER PRESENT?

		int icard = FitsJ.findCardWithKey ("SIMPLE",hdr);
		int jcard = FitsJ.findCardWithKey ("END",hdr);
		if (icard < 0 || jcard < 0) return;

		int naxis=0;
		hasRADEC = false;

		// GET NUMBER OF AXES

		naxis = -1;
		icard = FitsJ.findCardWithKey ("NAXIS",hdr);
		if (icard >= 0)
			NAXIS = FitsJ.getCardIntValue (hdr[icard]);
		if (NAXIS <= 0)
			{
			logInfo += "Can not read NAXIS keyword in FITS header of "+title+"\n";
			return;
			}
		naxis = NAXIS;

		// CHECK IF THE NUMBER OF WCS AXES IS KNOWN AND DIFFERENT

		icard = FitsJ.findCardWithKey ("WCSAXES",hdr);
		if (icard >= 0)
			{
			WCSAXES = FitsJ.getCardIntValue (hdr[icard]);
			naxis = WCSAXES;
			}

		// GIVEN NUMBER OF WCS AXES, RE-INITIALIZE MATRICES, ETC.

		initialize (naxis);

		// GET SIZES OF AXES

		for (int j=1; j <= naxis; j++)
			{
			icard = FitsJ.findCardWithKey ("NAXIS"+j, hdr);
			if (icard < 0)
				{
				logInfo += "Cannot find keyword NAXIS"+j+" in FITS header of "+title+"\n";
				return;
				}
			NAXES[j-1] = FitsJ.getCardIntValue (hdr[icard]);
			}

		// MAKE SURE THEY AGREE WITH THE ImageJ NUMBERS

		if (naxis >= 1 && nx >= 1 && NAXES[0] != nx)
			{
			logInfo += "Horizontal axis size="+nx+" does not match FITS header! ("+NAXES[0]+")\n";
			return;
			}
		if (naxis >= 2 && ny >= 1 && NAXES[1] != ny)
			{
			logInfo += "Vertial axis size="+ny+" does not match FITS header! ("+NAXES[1]+")\n";
			return;
			}
		if (naxis >= 3 && nz >= 1 && NAXES[2] != nz)
			{
			logInfo += "Stack size="+nz+" does not match FITS header! ("+NAXES[2]+")\n";
			return;
			}

		// TRY TO FIND RA-- and DEC- COORDINATE TYPES

		CTYPE[0] = null;
		String typ = "";
		String prefix = "";
		String abc[] = new String[] {	"",
						"A","B","C","D","E","F","G","H","I","J","K","L","M",
						"N","O","P","Q","R","S","T","U","V","W","X","Y","Z" };
		for (int k=0; k < abc.length && prefix == null; k++)
			{
			icard = FitsJ.findCardWithKey ("CTYPE1"+abc[k], hdr);
			if (icard > 0)
				{
				typ = FitsJ.getCardStringValue(hdr[icard]);
				if (typ != null && (typ.startsWith("RA--")))
					{
					CTYPE[0] = typ;
					prefix = abc[k];
					projection = CTYPE[0].substring(5, 8);
                    if (CTYPE[0].length() >= 12)
                        if (CTYPE[0].substring(9,12).equals("SIP"))
                            typeContainsSIP = true;
					break;
					}
				}
			}
		if (CTYPE[0] == null)	// NO "RA--" FOUND, SO TAKE WHAT YOU CAN GET
			{
			icard = FitsJ.findCardWithKey ("CTYPE1", hdr);
			if (icard > 0)
				{
				CTYPE[0] = FitsJ.getCardStringValue (hdr[icard]);
                projection = CTYPE[0].substring(5, 8);
                if (CTYPE[0].length() >= 12)
                    if (CTYPE[0].substring(9,12).equals("SIP"))
                        typeContainsSIP = true;
				}
			else	{
				logInfo += "Cannot define world coordinate system (CTYPE1): not enough information in FITS header!\n";
				return;
				}
			}

		// GET MATCHING COORDINATE TYPES

		for (int k=2; k <= naxis; k++)
			{
			icard = FitsJ.findCardWithKey ("CTYPE"+k+prefix, hdr);
			if (icard > 0)
				CTYPE[k-1] = FitsJ.getCardStringValue(hdr[icard]);
			}

		// CHECK IF CTYPE2n IS "DEC" AND COORDINATE SYSTEMS MATCH

		if (CTYPE[0].startsWith("RA--") && CTYPE[1].startsWith("DEC-"))
			hasRADEC=true;

		// GET TRANSFORMATION COEFFICIENTS (SCALE, TRANSLATION)

		try	{
			for (int k=1; k <= naxis; k++)
				{
				icard = FitsJ.findCardWithKey ("CDELT"+k+prefix, hdr);
				if (icard > 0)
					CDELT[k-1] = FitsJ.getCardDoubleValue(hdr[icard]);
				icard = FitsJ.findCardWithKey ("CRPIX"+k+prefix, hdr);
				if (icard > 0)
					CRPIX[k-1] = FitsJ.getCardDoubleValue(hdr[icard]);
				icard = FitsJ.findCardWithKey ("CRVAL"+k+prefix, hdr);
				if (icard > 0)
					CRVAL[k-1] = FitsJ.getCardDoubleValue(hdr[icard]);
				}
			}
		catch (NumberFormatException e)
			{
			logInfo += "Error reading standard transformation coefficients in FITS header!\n";
			return;
			}

		// GET LINEAR TRANSFORMATION COEFFICIENTS (ROTATION, SCALE, SKEWNESS)

		int ifound=0;
		try	{
			for (int k=1; k <= naxis; k++)
				{
				for (int i=1; i <= naxis; i++)
					{
					icard = FitsJ.findCardWithKey ("PC"+i+"_"+k+prefix, hdr);
					if (icard > 0)
						{
						PC[i-1][k-1] = FitsJ.getCardDoubleValue(hdr[icard]);
						ifound++;
						}
					icard = FitsJ.findCardWithKey ("CD"+i+"_"+k+prefix, hdr);
					if (icard > 0)
						{
						useCD=true;
						CD[i-1][k-1] = FitsJ.getCardDoubleValue(hdr[icard]);
						ifound++;
						}
					}
				}
			}
		catch (NumberFormatException e)
			{
			logInfo += "Cannot read transformation matrices PC/CD in FITS header!\n";
			return;
			}

		// IF AVAILABLE, GET SIP NON-LINEAR DISTORTION COEFFICIENTS (SEE SHUPE, D.L. ET AL. 2008)
        if (typeContainsSIP)
            {
            try	{
                // LOOK FOR A AND B POLYNOMIAL ORDERS
                if (naxis > 1)
                    {
                    A_ORDER = -1;
                    icard = FitsJ.findCardWithKey ("A_ORDER",hdr);
                    if (icard >= 0)
                        {
                        A_ORDER = FitsJ.getCardIntValue (hdr[icard]);
                        if (A_ORDER < 2 || A_ORDER > 9)
                            {
                            A_ORDER = -1;
                            logInfo += "SIP A_ORDER out of range in FITS header!\n";
                            }
                        }
                    else
                        logInfo += "SIP A_ORDER not found in FITS header!\n";

                    B_ORDER = -1;
                    icard = FitsJ.findCardWithKey ("B_ORDER",hdr);
                    if (icard >= 0)
                        {
                        B_ORDER = FitsJ.getCardIntValue (hdr[icard]);
                        if (B_ORDER < 2 || B_ORDER > 9)
                            {
                            B_ORDER = -1;
                            logInfo += "SIP B_ORDER out of range in FITS header!\n";
                            }
                        }
                    else
                        logInfo += "SIP B_ORDER not found in FITS header!\n";
                    }
                }
            catch (NumberFormatException e)
                {
                logInfo += "Cannot read SIP A_ORDER AND/OR B_ORDER in FITS header!\n";
                A_ORDER = -1;
                B_ORDER = -1;
                }

            //INITIALIZE A AND B MATRICES
            if (A_ORDER >= 2 && A_ORDER <= 9)
                {
                A = new double[A_ORDER+1][A_ORDER+1];
                for (int q=0; q <= A_ORDER; q++)
                    {
                    for (int p=0; p <= A_ORDER; p++)
                        {
                        A[p][q] = 0.0;
                        }
                    }
                }

            if (B_ORDER >= 2 && B_ORDER <= 9)
                {
                B = new double[B_ORDER+1][B_ORDER+1];
                for (int q=0; q <= B_ORDER; q++)
                    {
                    for (int p=0; p <= B_ORDER; p++)
                        {
                        B[p][q] = 0.0;
                        }
                    }
                }

            //LOOK FOR SIP X-COORDINATE MATRIX COEFICIENTS A_p_q IN FITS HEADER
            if (A_ORDER >= 2 && A_ORDER <= 9)
                {
                try	{
                    for (int q=0; q <= A_ORDER; q++)
                        {
                        for (int p=0; p <= A_ORDER; p++)
                            {
                            if (q + p <= A_ORDER)
                                {
                                icard = FitsJ.findCardWithKey ("A_"+p+"_"+q+prefix, hdr);
                                if (icard > 0)
                                    {
                                    A[p][q] = FitsJ.getCardDoubleValue(hdr[icard]);
//                                    IJ.write("A["+p+"]["+q+"] = "+A[p][q]);
                                    useSIPA=true;
                                    }
                                }
                            }
                        }
                    }
                catch (NumberFormatException e)
                    {
                    logInfo += "Cannot read SIP matrix A in FITS header!\n";
                    return;
                    }
                }
            
            //LOOK FOR SIP Y-COORDINATE MATRIX COEFICIENTS B_p_q IN FITS HEADER
            if (B_ORDER >= 2 && B_ORDER <= 9)
                {
                try	{
                    for (int q=0; q <= B_ORDER; q++)
                        {
                        for (int p=0; p <= B_ORDER; p++)
                            {
                            if (q + p <= B_ORDER)
                                {
                                icard = FitsJ.findCardWithKey ("B_"+p+"_"+q+prefix, hdr);
                                if (icard > 0)
                                    {
                                    B[p][q] = FitsJ.getCardDoubleValue(hdr[icard]);
//                                    IJ.write("B["+p+"]["+q+"] = "+B[p][q]);
                                    useSIPB=true;
                                    }
                                }
                            }
                        }
                    }
                catch (NumberFormatException e)
                    {
                    logInfo += "Cannot read SIP matrix B in FITS header!\n";
                    return;
                    }
                }
            }
		// GET PROJECTION PARAMETERS, IF ANY

		LONPOLE = 180.0;// DEGS
		icard = FitsJ.findCardWithKey ("LONPOLE"+prefix, hdr);
		if (icard > 0)
			LONPOLE = FitsJ.getCardDoubleValue(hdr[icard]);

		// COORDINATE SYSTEM

		coordsys = CTYPE[0].trim()+","+CTYPE[1].trim();
		enoughInfo = true;
		return;
		}

	public void initialize (int naxis)
		{
		NAXIS = naxis;
		if (naxis == 0) return;

		NAXES = new int[NAXIS];
		CDELT = new double[NAXIS];
		CRPIX = new double[NAXIS];
		CRVAL = new double[NAXIS];
		PC = new double[NAXIS][NAXIS];
		CD = new double[NAXIS][NAXIS];
		CTYPE = new String[NAXIS];
		CUNIT = new String[NAXIS];
		for (int i=0; i < NAXIS; i++)
			{
			NAXES[i]=0;
			CDELT[i]=1.0;
			CRPIX[i]=1.0;
			CRVAL[i]=0.0;
			PC[i][i]=1.0;
			CD[i][i]=1.0;
			}
        typeContainsSIP=false;
		useCD=false;
        useSIPA=false;
        useSIPB=false;
		coordsys = "";
		projection = "";
		}

	/**
	 * Converts ImageJ pixel positions to coordinate system values.  Note that ImageJ pixels are assumed to
	 * be shifted by Centroid.PIXELCENTER pixels relative to the FITS standard.
	 */
	public double[] pixels2wcs (double[] pixels)
		{
		if (! enoughInfo)
			{
			IJ.beep();
			IJ.log ("Not enough info: "+logInfo);
			return null;
			}

		int n=pixels.length;
		if (n != NAXIS)
			{
			IJ.beep();
			IJ.log ("Number of axes doesn't match: n="+n+", NAXIS="+NAXIS);
			return null;
			}

		// TRANSLATE IJ COORDINATES TO FITS COORDINATES
		double[] p = new double[n];	// IN PIXELS
		for (int j=0; j < n; j++)
			{
			if (j == 1)
				p[1] = NAXES[1] - (pixels[1] - Centroid.PIXELCENTER) - CRPIX[1]; 	
			else
				p[j] = (pixels[j] - Centroid.PIXELCENTER) + 1.0 - CRPIX[j];		
			}

        // NON-LINEAR DISTORTION CORRECTION USING SIP CORRECTION MATRICES (SEE SHUPE, D.L. ET AL. 2008)
        double p0 = p[0];
        if (useSIPA)
            {
            //CALCULATE SIP X-COORDINATE CORRECTION
            double xCorrection = 0.0;
			for (int qq=0; qq <= A_ORDER; qq++)
				{
				for (int pp=0; pp <= A_ORDER; pp++)
					{
                    if (pp + qq <= A_ORDER)
                        {
                        xCorrection += A[pp][qq]*Math.pow(p0,(double)pp)*Math.pow(p[1],(double)qq);
                        }
					}
				}
            p[0] += xCorrection;
            }

        if (useSIPB)
            {
            //CALCULATE SIP Y-COORDINATE CORRECTION
            double yCorrection = 0.0;
			for (int qq=0; qq <= B_ORDER; qq++)
				{
				for (int pp=0; pp <= B_ORDER; pp++)
					{
                    if (pp + qq <= B_ORDER)
                        {
                        yCorrection += B[pp][qq]*Math.pow(p0,(double)pp)*Math.pow(p[1],(double)qq);
                        }
					}
				}
            p[1] += yCorrection;
            }

		// CORRECT FOR ROTATION, SKEWNESS, SCALE

        double[] x = new double[n];	// IN PSUEDO DEGREES
		if (useCD)
			{
			for (int j=0; j < n; j++)
				{
				for (int i=0; i < n; i++)
                    x[j] += CD[j][i]*p[i];
				}
			}
		else	{
			for (int j=0; j < n; j++)
				{
				for (int i=0; i < n; i++)
					x[j] += PC[j][i]*p[i];
				x[j] *= CDELT[j];
				}
			}

		// PROJECTION PLANE COORDINATES x TO NATIVE SPHERICAL COORDINATES s
		double[] s = new double[n];	// IN PSUEDO DEGREES
		double DEGRAD = 180.0/Math.PI;
		for (int j=0; j < n; j++)
			s[j] = x[j];
		if (projection.equals("TAN"))
			{
			double Rtheta = Math.sqrt(x[0]*x[0]+x[1]*x[1]);
			s[0] = DEGRAD*Math.atan2(x[0],-x[1]);		    // NATIVE phi (E.G. R.A.) IN DEGS
			s[1] = DEGRAD*Math.atan2(DEGRAD,Rtheta);		// NATIVE theta~90 DEG (E.G. DEC.) IN DEGS
			}

		// NATIVE SPHERICAL COORDINATES s TO CELESTIAL COORDINATES c : SEE EQN.2 OF Calabretta & Greisen 2002, P. 1079
		double[] c = new double[n];
		for (int j=0; j < n; j++)
			c[j] = s[j]+CRVAL[j];
		if (!projection.equals(""))
			{
			double phi = s[0]/DEGRAD;		// NATIVE PHI IN RADIANS
			double theta = s[1]/DEGRAD;		// NATIVE THETA IN RADIANS
			double phiPole = LONPOLE /DEGRAD;	// IN RADIANS
			double deltaPole = CRVAL[1]/DEGRAD;	// IN RADIANS
			double sintcosP = Math.sin(theta)*Math.cos(deltaPole);
			double costsinPcosdp = Math.cos(theta)*Math.sin(deltaPole)*Math.cos(phi-phiPole);
			double costsindp = Math.cos(theta)*Math.sin(phi-phiPole);
			double sintsinP = Math.sin(theta)*Math.sin(deltaPole);
			double costcosPcosdp = Math.cos(theta)*Math.cos(deltaPole)*Math.cos(phi-phiPole);
			c[0] = CRVAL[0]+Math.atan2(-costsindp, sintcosP-costsinPcosdp)*DEGRAD;
			c[1] = Math.asin(sintsinP+costcosPcosdp)*DEGRAD;
			}
		return c;
		}

	public boolean hasRaDec ()
		{
		return hasRADEC;
		}
	public boolean hasWCS ()
		{
		return enoughInfo;
		}

	public void repair (ImagePlus img, double[] crpix, double[][] cd)
		{
		NAXIS=0;
		String[] hdr = FitsJ.getHeader (img);
		if (hdr == null) return;

		// FIX CTYPEn KEYWORDS
		int icard=FitsJ.findCardWithKey("CTYPE1",hdr);
		int jcard=FitsJ.findCardWithKey("CTYPE2",hdr);
		if (icard < 0 && jcard < 0)
			{
			hdr = FitsJ.setCard("CTYPE1","RA---TAN","coordinate type (WCS repair)",hdr);
			hdr = FitsJ.setCard("CTYPE2","DEC--TAN","coordinate type (WCS repair)",hdr);
			}

		// FIX CRVALn
		icard=FitsJ.findCardWithKey("CRVAL1",hdr);
		jcard=FitsJ.findCardWithKey("CRVAL2",hdr);
		if (icard < 0 && jcard < 0)
			{
			icard=FitsJ.findCardWithKey("RA",hdr);
			jcard=FitsJ.findCardWithKey("DEC",hdr);
			if (icard > 0 && jcard > 0)
				{
				try	{
					double ra = DmsFormat.unformat(FitsJ.getCardStringValue(hdr[icard]));
					double dec = DmsFormat.unformat(FitsJ.getCardStringValue(hdr[jcard]));
					hdr = FitsJ.setCard("CRVAL1",ra*15.0,"Right Ascension in decimal degrees (WCS repair)",hdr);
					hdr = FitsJ.setCard("CRVAL2",dec,"Declination in decimal degrees (WCS repair)",hdr);
					}
				catch (NumberFormatException e)
					{
					IJ.error("Cannot parse RA, DEC entries in FITS header!");
					return;
					}
				}
			else	{
				IJ.error("Cannot find RA,DEC in FITS header!");
				return;
				}
			}
		
		// FIX CRPIXn
		icard=FitsJ.findCardWithKey("CRPIX1",hdr);
		jcard=FitsJ.findCardWithKey("CRPIX2",hdr);
		if (icard < 0 && jcard < 0)
			{
			hdr = FitsJ.setCard("CRPIX1",crpix[0],"reference pixel coordinate (WCS repair)",hdr);
			hdr = FitsJ.setCard("CRPIX2",crpix[1],"reference pixel coordinate (WCS repair)",hdr);
			}

		// FIX CDij
		icard=FitsJ.findCardWithKey("CD1_1",hdr);
		if (icard < 0)
			{
			hdr = FitsJ.setCard("CD1_1",cd[0][0],"transformation matrix element (WCS repair)",hdr);
			hdr = FitsJ.setCard("CD1_2",cd[0][1],"transformation matrix element (WCS repair)",hdr);
			hdr = FitsJ.setCard("CD2_1",cd[1][0],"transformation matrix element (WCS repair)",hdr);
			hdr = FitsJ.setCard("CD2_2",cd[1][1],"transformation matrix element (WCS repair)",hdr);
			}

		// SAVE RESULT
		FitsJ.putHeader(img,hdr);

		// TRY AGAIN
		process(img.getShortTitle(),img.getWidth(),img.getHeight(),img.getStackSize(),hdr);
		}
	}
